==> wyvern-2.2.7.tar.bz2: wyvern-2.2.7/modules/mod_access_sqlite.c <==
/*****************************************************************************
 *
 * FILE:	mod_access_sqlite.c
 * DESCRIPTION:	Wyvern: access control module with SQLite
 * DATE:	Sat, Feb 18 2006
 * UPDATED:	Wed, Mar  1 2006
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_access_sqlite.c,v 1.3 2006/03/01 13:21:37 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <sqlite3.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_access_sqlite"

typedef struct {
  char *		path;	/* path of database file */
  sqlite3 *		sdb;
  sqlite3_stmt *	stmt;
} wmDBI;

#define	WYVERN_ACCESS_SQLITE	"sqlite_access.db"

#if	CONTINUOUS_CONNECTION
#define	SELECT_ACCESS \
	"SELECT client,access FROM wyvern_access" \
 	"  WHERE (server = $NAME OR server = 'any')" \
 	"  AND dirname = substr($URI,0,length(dirname));"
#else
#define	SELECT_ACCESS \
	"SELECT client,access FROM wyvern_access" \
 	"  WHERE (server = '%q' OR server = 'any')" \
 	"  AND dirname = substr('%q',0,length(dirname));"
#endif	/* CONTINUOUS_CONNECTION */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	wm_access_sqlite_init(void);
static void	wm_access_sqlite_finish(void);
static int	wm_access_sqlite_checker(Message *);
static int	access_checker(Message *, const char *);
static int	compare_domain(const char *, const char *);
static int	compare_address(const char *, const char *);
static void	wm_access_sqlite_register_hooks(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static wmDBI	dbi = { NULL, NULL, NULL };

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_access_sqlite_init
 * Description:		initialize the host access list
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_access_sqlite_init(void)
{
#if	CONTINUOUS_CONNECTION
  int	rc;
#endif	/* CONTINUOUS_CONNECTION */

  /* read 'ServerRoot/conf/waccess.conf' */
  ws_asprintf(&dbi.path, "%s/%s", httpd->configDir, WYVERN_ACCESS_SQLITE);
#if	CONTINUOUS_CONNECTION
  rc = sqlite3_open(dbi.path, &dbi.sdb);
  if (rc != SQLITE_OK) {
    wv_die("[%s] failed to open: %s", WM_MODULE_ID, dbi.path);
  }
  rc = sqlite3_prepare(dbi.sdb, SELECT_ACCESS, (int)strlen(SELECT_ACCESS),
		       &dbi.stmt, NULL);
  if (rc != SQLITE_OK) {
    wv_die("[%s] failed to prepare: '%s'", WM_MODULE_ID, SELECT_ACCESS);
  }
#endif	/* CONTINUOUS_CONNECTION */
}

/*
 * Function Name:	wm_access_sqlite_finish
 * Description:		release the memory area of access lists
 * Arguments:		none
 * Return Value:	none
 */
void
wm_access_sqlite_finish(void)
{
  wv_free(dbi.path);
#if	CONTINUOUS_CONNECTION
  if (dbi.stmt != NULL) { sqlite3_finalize(dbi.stmt); dbi.stmt = NULL; }
  if (dbi.sdb  != NULL) { sqlite3_close(dbi.sdb);     dbi.sdb  = NULL; }
#endif	/* CONTINUOUS_CONNECTION */
}

/*
 * Function Name:	wm_access_sqlite_checker
 * Description:		permit the access to the host
 * Arguments:		message
 * Return Value:	allowed host: 0, denied host: -1
 */
static int
wm_access_sqlite_checker(msg)
	register Message *	msg;
{
  int	rc;
  int	retval = WM_RET_OK;

#if	CONTINUOUS_CONNECTION
  sqlite3_bind_text(dbi.stmt, 1, msg->vhost->serverName,
		    (int)strlen(msg->vhost->serverName), SQLITE_TRANSIENT);
  sqlite3_bind_text(dbi.stmt, 2, msg->request.uri,
		    (int)strlen(msg->request.uri), SQLITE_TRANSIENT);
  do {
    rc = sqlite3_step(dbi.stmt);
    if (rc == SQLITE_ROW && sqlite3_data_count(dbi.stmt) == 2) {
      const char *	client;
      const char *	access;

      client = sqlite3_column_text(dbi.stmt, 0);
      access = sqlite3_column_text(dbi.stmt, 1);

#if	DEBUG
      fprintf(stderr, "DEBUG[%s] %s => %s\n",
	      WM_MODULE_ID, client, access);
#endif	/* DEBUG */
      if (access_checker(msg, client) == 0) {
	retval = (ws_strcmp(access, "deny") == 0) ? WM_RET_ERROR : WM_RET_OK;
	break;	/* matched */
      }
    }
  } while (rc == SQLITE_ROW);
  sqlite3_reset(dbi.stmt);
#else
  rc = sqlite3_open(dbi.path, &dbi.sdb);
  if (rc == SQLITE_OK) {
    char *	sql;

    sql = sqlite3_mprintf(SELECT_ACCESS,
			  msg->vhost->serverName, msg->request.uri);
    if (sql != NULL) {
      sqlite3_stmt *	stmt;
      const char *	tail;

      rc = sqlite3_prepare(dbi.sdb, sql, (int)strlen(sql), &stmt, &tail);
      if (rc == SQLITE_OK) {
	do {
	  rc = sqlite3_step(stmt);
	  if (rc == SQLITE_ROW && sqlite3_data_count(stmt) == 2) {
	    const char *	client;
	    const char *	access;

	    client = sqlite3_column_text(stmt, 0);
	    access = sqlite3_column_text(stmt, 1);

#if	DEBUG
	    fprintf(stderr, "DEBUG[%s] %s => %s\n",
		    WM_MODULE_ID, client, access);
#endif	/* DEBUG */
	    if (access_checker(msg, client) == 0) {
	      retval =  (ws_strcmp(access, "deny") == 0)
		     ? WM_RET_ERROR
		     : WM_RET_OK;
	      break;	/* matched */
	    }
	  }
	} while (rc == SQLITE_ROW);
	sqlite3_finalize(stmt);
      }
      sqlite3_free(sql);
    }
    sqlite3_close(dbi.sdb);
  }
#endif	/* CONTINUOUS_CONNECTION */

  return retval;
}

static int
access_checker(msg, client)
	register Message *	msg;
	const char *		client;
{
  struct sockaddr	sa;
  char			hbuf[NI_MAXHOST];

  DBG(fputs("<access_checker>\n", stderr);)

  /* get client hostname with DNS */
  memset(hbuf, 0, sizeof(hbuf));
  memcpy(&sa, msg->client.addr, sizeof(struct sockaddr));
#if	HAVE_SOCKADDR_SA_LEN
  if (getnameinfo(&sa, sa.sa_len, hbuf, sizeof(hbuf), NULL, 0,
		  NI_NAMEREQD) != 0) {
    wv_syslog(WLL_LOG_SECURITY, "[%s] No canonical name: %s",
	      WM_MODULE_ID, msg->client.ip);
    return -1;
  }
#else
  if (getnameinfo(&sa, sizeof(sa), hbuf, sizeof(hbuf), NULL, 0,
		  NI_NAMEREQD) != 0) {
    wv_syslog(WLL_LOG_SECURITY, "[%s] No canonical name: %s",
	      WM_MODULE_ID, msg->client.ip);
    return -1;
  }
#endif	/* HAVE_SOCKADDR_SA_LEN */
  msg->client.name = ws_strdup(hbuf);

  if (compare_domain(msg->client.name, client) == 0) {
    return 0;	/* matched */
  }
  if (compare_address(msg->client.ip, client) == 0) {
    return 0;	/* matched */
  }

  return -1;
}

static int
compare_domain(name, pat)
	register const char *	name;
	register const char *	pat;
{
  size_t	nlen;
  size_t	plen;

  if (name == NULL || pat == NULL) {
    return -1;
  }

  DBG(fprintf(stderr, "[compare_domain] %s<=>%s\n", name, pat);)
  nlen = strlen(name);
  plen = strlen(pat);
  if (nlen < plen) {
    return -1;
  }

  /* check to match the name with the pattern like as '.ad.jp' */
  if (nlen >= plen && pat[0] == '.' &&
      ws_strncmp(name + nlen - plen, pat, plen) == 0) {
    return 0;	/* match! */
  }
  /* check to match the name exactly */
  if (ws_strcmp(name, pat) == 0) {
    return 0;	/* match! */
  }

  return -1;	/* no match */
}

/*****************************************************************************/

typedef union {
  in_addr_t	bits;
  uint8_t	a[4];
} wv_netmask_t;

typedef struct {
  uint8_t	inet[4];
  wv_netmask_t	mask;
} wv_ipaddr_t;

static int	parse_addr(const char *, wv_ipaddr_t *);
static int	parse_cidr(const char *, wv_ipaddr_t **);
static int	check_addr(const char *, wv_ipaddr_t *);

static int
parse_addr(ipv4, retval)
	const char *	ipv4;
	wv_ipaddr_t *	retval;
{
  char *	s;
  char **	ap;
  char *	args[4];
  int		n;
  register int	i;
  wv_netmask_t	x;

  s = ws_strdup(ipv4);

  for (n = 0, ap = args; (*ap = strsep(&s, ".")) != NULL; ) {
    if (**ap != '\0') {
      n++;
      if (++ap >= &args[4]) {
	break;
      }
    }
  }

  x.bits = htonl(retval->mask.bits);
  for (i = 0; i < n; i++) {
    retval->inet[i] = (uint8_t)strtol(args[i], (char **)NULL, 10) & x.a[i];
  }
  free(s);

  return 0;
}

static int
parse_cidr(ipv4, retval)
	const char *	ipv4;
	wv_ipaddr_t **	retval;
{
  char *		ap;
  register char *	cp;
  int			cidr = 32;
  in_addr_t		mask = 0xffffffff;

  ap = ws_strdup(ipv4);
  cp = strchr(ap, '/');
  if (cp != NULL) {
    *cp++ = '\0';
    cidr = (int)strtol(cp, (char **)NULL, 10);
  }
  mask <<= (32 - cidr);

  if (retval != NULL) {
    *retval		 = (wv_ipaddr_t *)ws_malloc(sizeof(wv_ipaddr_t));
    (*retval)->mask.bits = mask;
    if (parse_addr(ap, *retval) == -1) {
      free(*retval);
      *retval = NULL;
    }
  }
  free(ap);

  return (*retval != NULL) ? 0 : -1;
}

static int
check_addr(addr, net)
	const char *	addr;
	wv_ipaddr_t *	net;
{
  wv_ipaddr_t *	ip;
  int		retval = 0;

  if (parse_cidr(addr, &ip) == 0) {
    register int	i;
    wv_netmask_t	x;

    x.bits = htonl(net->mask.bits);
    for (i = 0; i < 4; i++) {
      if (net->inet[i] != (ip->inet[i] & x.a[i])) {
	retval = -1;
      }
    }
    free(ip);
  }

  return retval;
}

static int
compare_address(name, pat)
	register const char *	name;
	register const char *	pat;
{
  wv_ipaddr_t *	addr;
  int		retval = -1;

  DBG(fprintf(stderr, "[compare_address] %s<=>%s\n", name, pat);)
  if (parse_cidr(pat, &addr) == 0) {
    retval = check_addr(name, addr);
#if	DEBUG
    fprintf(stderr, "[compare_address] %s<=>%d.%d.%d.%d/%x [%d]\n",
	    name,
	    addr->inet[0], addr->inet[1], addr->inet[2], addr->inet[3],
	    addr->mask.bits, retval);
#endif	/* DEBUG */
    free(addr);
  }

  return retval;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_access_sqlite_register_hooks(void)
{
  wm_hook_access_checker(wm_access_sqlite_checker, HOOK_MIDDLE);
}

static wmModule	mod_access_sqlite = {
  WM_MAGIC_NUMBER_MAJOR,			/* version */
  WM_SECURE,					/* secure */
  WM_MODULE_ID,					/* name */
  "Access Control Module with SQLite version 1.0.1",	/* info */
  "Kouichi ABE (WALL)",				/* author */
  NULL,						/* depend */
  NULL,						/* handle */
  (void *)wm_access_sqlite_init,		/* init */
  (void *)wm_access_sqlite_finish,		/* finish */
  (void *)wm_access_sqlite_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_access_sqlite;
}

==> wyvern-2.2.10.tar.bz2: wyvern-2.2.10/modules/mod_access_sqlite.c <==
/*****************************************************************************
 *
 * FILE:	mod_access_sqlite.c
 * DESCRIPTION:	Wyvern: access control module with SQLite
 * DATE:	Sat, Feb 18 2006
 * UPDATED:	Sat, Aug 25 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2006-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2006-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_access_sqlite.c,v 1.4 2007/08/24 18:01:55 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#include <sqlite3.h>
#include "wyvern.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	WM_MODULE_ID	"mod_access_sqlite"

typedef struct {
  char *		path;	/* path of database file */
  sqlite3 *		sdb;
  sqlite3_stmt *	stmt;
} wmDBI;

#define	WYVERN_ACCESS_SQLITE	"sqlite_access.db"

#if	CONTINUOUS_CONNECTION
#define	SELECT_ACCESS \
	"SELECT client,access FROM wyvern_access" \
 	"  WHERE (server = $NAME OR server = 'any')" \
 	"  AND dirname = substr($URI,0,length(dirname));"
#else
#define	SELECT_ACCESS \
	"SELECT client,access FROM wyvern_access" \
 	"  WHERE (server = '%q' OR server = 'any')" \
 	"  AND dirname = substr('%q',0,length(dirname));"
#endif	/* CONTINUOUS_CONNECTION */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	wm_access_sqlite_init(void);
static void	wm_access_sqlite_finish(void);
static int	wm_access_sqlite_checker(Message *);
static int	access_checker(Message *, const char *);
static int	compare_domain(const char *, const char *);
static int	compare_address(const char *, const char *);
static void	wm_access_sqlite_register_hooks(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static wmDBI	dbi = { NULL, NULL, NULL };

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_access_sqlite_init
 * Description:		initialize the host access list
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_access_sqlite_init(void)
{
#if	CONTINUOUS_CONNECTION
  int	rc;
#endif	/* CONTINUOUS_CONNECTION */

  /* read 'ServerRoot/conf/waccess.conf' */
  ws_asprintf(&dbi.path, "%s/%s", httpd->configDir, WYVERN_ACCESS_SQLITE);
#if	CONTINUOUS_CONNECTION
  rc = sqlite3_open(dbi.path, &dbi.sdb);
  if (rc != SQLITE_OK) {
    wv_die("[%s] failed to open: %s", WM_MODULE_ID, dbi.path);
  }
  rc = sqlite3_prepare(dbi.sdb, SELECT_ACCESS, (int)strlen(SELECT_ACCESS),
		       &dbi.stmt, NULL);
  if (rc != SQLITE_OK) {
    wv_die("[%s] failed to prepare: '%s'", WM_MODULE_ID, SELECT_ACCESS);
  }
#endif	/* CONTINUOUS_CONNECTION */
}

/*
 * Function Name:	wm_access_sqlite_finish
 * Description:		release the memory area of access lists
 * Arguments:		none
 * Return Value:	none
 */
void
wm_access_sqlite_finish(void)
{
  wv_free(dbi.path);
#if	CONTINUOUS_CONNECTION
  if (dbi.stmt != NULL) { sqlite3_finalize(dbi.stmt); dbi.stmt = NULL; }
  if (dbi.sdb  != NULL) { sqlite3_close(dbi.sdb);     dbi.sdb  = NULL; }
#endif	/* CONTINUOUS_CONNECTION */
}

/*
 * Function Name:	wm_access_sqlite_checker
 * Description:		permit the access to the host
 * Arguments:		message
 * Return Value:	allowed host: 0, denied host: -1
 */
static int
wm_access_sqlite_checker(msg)
	register Message *	msg;
{
  int	rc;
  int	status = WM_RET_OK;

#if	CONTINUOUS_CONNECTION
  sqlite3_bind_text(dbi.stmt, 1, msg->vhost->serverName,
		    (int)strlen(msg->vhost->serverName), SQLITE_TRANSIENT);
  sqlite3_bind_text(dbi.stmt, 2, msg->request.uri,
		    (int)strlen(msg->request.uri), SQLITE_TRANSIENT);
  do {
    rc = sqlite3_step(dbi.stmt);
    if (rc == SQLITE_ROW && sqlite3_data_count(dbi.stmt) == 2) {
      const char *	client;
      const char *	access;

      client = sqlite3_column_text(dbi.stmt, 0);
      access = sqlite3_column_text(dbi.stmt, 1);

#if	DEBUG
      fprintf(stderr, "DEBUG[%s] %s => %s\n",
	      WM_MODULE_ID, client, access);
#endif	/* DEBUG */
      if (access_checker(msg, client) == 0) {
	status = (ws_strcmp(access, "deny") == 0) ? WM_RET_ERROR : WM_RET_OK;
	break;	/* matched */
      }
    }
  } while (rc == SQLITE_ROW);
  sqlite3_reset(dbi.stmt);
#else
  rc = sqlite3_open(dbi.path, &dbi.sdb);
  if (rc == SQLITE_OK) {
    char *	sql;

    sql = sqlite3_mprintf(SELECT_ACCESS,
			  msg->vhost->serverName, msg->request.uri);
    if (sql != NULL) {
      sqlite3_stmt *	stmt;
      const char *	tail;

      rc = sqlite3_prepare(dbi.sdb, sql, (int)strlen(sql), &stmt, &tail);
      if (rc == SQLITE_OK) {
	do {
	  rc = sqlite3_step(stmt);
	  if (rc == SQLITE_ROW && sqlite3_data_count(stmt) == 2) {
	    const char *	client;
	    const char *	access;

	    client = sqlite3_column_text(stmt, 0);
	    access = sqlite3_column_text(stmt, 1);

#if	DEBUG
	    fprintf(stderr, "DEBUG[%s] %s => %s\n",
		    WM_MODULE_ID, client, access);
#endif	/* DEBUG */
	    if (access_checker(msg, client) == 0) {
	      status =  (ws_strcmp(access, "deny") == 0)
		     ? WM_RET_ERROR
		     : WM_RET_OK;
	      break;	/* matched */
	    }
	  }
	} while (rc == SQLITE_ROW);
	sqlite3_finalize(stmt);
      }
      sqlite3_free(sql);
    }
    sqlite3_close(dbi.sdb);
  }
#endif	/* CONTINUOUS_CONNECTION */

  return status;
}

static int
access_checker(msg, client)
	register Message *	msg;
	const char *		client;
{
  struct sockaddr	sa;
  char			hbuf[NI_MAXHOST];

  DBG(fputs("<access_checker>\n", stderr);)

  /* get client hostname with DNS */
  memset(hbuf, 0, sizeof(hbuf));
  memcpy(&sa, msg->client.addr, sizeof(struct sockaddr));
#if	HAVE_SOCKADDR_SA_LEN
  if (getnameinfo(&sa, sa.sa_len, hbuf, sizeof(hbuf), NULL, 0,
		  NI_NAMEREQD) != 0) {
    wv_syslog(WLL_LOG_SECURITY, "[%s] No canonical name: %s",
	      WM_MODULE_ID, msg->client.ip);
    return -1;
  }
#else
  if (getnameinfo(&sa, sizeof(sa), hbuf, sizeof(hbuf), NULL, 0,
		  NI_NAMEREQD) != 0) {
    wv_syslog(WLL_LOG_SECURITY, "[%s] No canonical name: %s",
	      WM_MODULE_ID, msg->client.ip);
    return -1;
  }
#endif	/* HAVE_SOCKADDR_SA_LEN */
  msg->client.name = ws_strdup(hbuf);

  if (compare_domain(msg->client.name, client) == 0) {
    return 0;	/* matched */
  }
  if (compare_address(msg->client.ip, client) == 0) {
    return 0;	/* matched */
  }

  return -1;
}

static int
compare_domain(name, pat)
	register const char *	name;
	register const char *	pat;
{
  size_t	nlen;
  size_t	plen;

  if (name == NULL || pat == NULL) {
    return -1;
  }

  DBG(fprintf(stderr, "[compare_domain] %s<=>%s\n", name, pat);)
  nlen = strlen(name);
  plen = strlen(pat);
  if (nlen < plen) {
    return -1;
  }

  /* check to match the name with the pattern like as '.ad.jp' */
  if (nlen >= plen && pat[0] == '.' &&
      ws_strncmp(name + nlen - plen, pat, plen) == 0) {
    return 0;	/* match! */
  }
  /* check to match the name exactly */
  if (ws_strcmp(name, pat) == 0) {
    return 0;	/* match! */
  }

  return -1;	/* no match */
}

/*****************************************************************************/

typedef union {
  in_addr_t	bits;
  uint8_t	a[4];
} wv_netmask_t;

typedef struct {
  uint8_t	inet[4];
  wv_netmask_t	mask;
} wv_ipaddr_t;

static int	parse_addr(const char *, wv_ipaddr_t *);
static int	parse_cidr(const char *, wv_ipaddr_t **);
static int	check_addr(const char *, wv_ipaddr_t *);

static int
parse_addr(ipv4, retval)
	const char *	ipv4;
	wv_ipaddr_t *	retval;
{
  char *	s;
  char **	ap;
  char *	args[4];
  int		n;
  register int	i;
  wv_netmask_t	x;

  s = ws_strdup(ipv4);

  for (n = 0, ap = args; (*ap = strsep(&s, ".")) != NULL; ) {
    if (**ap != '\0') {
      n++;
      if (++ap >= &args[4]) {
	break;
      }
    }
  }

  x.bits = htonl(retval->mask.bits);
  for (i = 0; i < n; i++) {
    retval->inet[i] = (uint8_t)strtol(args[i], (char **)NULL, 10) & x.a[i];
  }
  free(s);

  return 0;
}

static int
parse_cidr(ipv4, retval)
	const char *	ipv4;
	wv_ipaddr_t **	retval;
{
  char *		ap;
  register char *	cp;
  int			cidr = 32;
  in_addr_t		mask = 0xffffffff;

  ap = ws_strdup(ipv4);
  cp = strchr(ap, '/');
  if (cp != NULL) {
    *cp++ = '\0';
    cidr = (int)strtol(cp, (char **)NULL, 10);
  }
  mask <<= (32 - cidr);

  if (retval != NULL) {
    *retval		 = (wv_ipaddr_t *)ws_malloc(sizeof(wv_ipaddr_t));
    (*retval)->mask.bits = mask;
    if (parse_addr(ap, *retval) == -1) {
      free(*retval);
      *retval = NULL;
    }
  }
  free(ap);

  return (*retval != NULL) ? 0 : -1;
}

static int
check_addr(addr, net)
	const char *	addr;
	wv_ipaddr_t *	net;
{
  wv_ipaddr_t *	ip;
  int		status = 0;

  if (parse_cidr(addr, &ip) == 0) {
    register int	i;
    wv_netmask_t	x;

    x.bits = htonl(net->mask.bits);
    for (i = 0; i < 4; i++) {
      if (net->inet[i] != (ip->inet[i] & x.a[i])) {
	status = -1;
      }
    }
    free(ip);
  }

  return status;
}

static int
compare_address(name, pat)
	register const char *	name;
	register const char *	pat;
{
  wv_ipaddr_t *	addr;
  int		status = -1;

  DBG(fprintf(stderr, "[compare_address] %s<=>%s\n", name, pat);)
  if (parse_cidr(pat, &addr) == 0) {
    status = check_addr(name, addr);
#if	DEBUG
    fprintf(stderr, "[compare_address] %s<=>%d.%d.%d.%d/%x [%d]\n",
	    name,
	    addr->inet[0], addr->inet[1], addr->inet[2], addr->inet[3],
	    addr->mask.bits, status);
#endif	/* DEBUG */
    free(addr);
  }

  return status;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_access_sqlite_register_hooks(void)
{
  wm_hook_access_checker(wm_access_sqlite_checker, HOOK_MIDDLE);
}

static wmModule	mod_access_sqlite = {
  WM_MAGIC_NUMBER_MAJOR,			/* version */
  WM_SECURE,					/* secure */
  WM_MODULE_ID,					/* name */
  "Access Control Module with SQLite version 1.0.1",	/* info */
  "Kouichi ABE (WALL)",				/* author */
  NULL,						/* depend */
  NULL,						/* handle */
  (void *)wm_access_sqlite_init,		/* init */
  (void *)wm_access_sqlite_finish,		/* finish */
  (void *)wm_access_sqlite_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_access_sqlite;
}
